/* fraglets0.18/src/frag.c

  fraglet interpreter

  Copyright (C) 2003-2005 <christian.tschudin@unibas.ch>

  This file is distributed under the terms of the GNU general Public
  License (GPL), see the file COPYING

  contains the generic fraglet interpreter
  as well (at the end) all the problem specific
  routines and fitness function support (score).

  History:

  20050220  lidia.yamamoto@unibas.ch: reset random seed for repeatability
  20050217  christian.tschudin@unibas.ch (fraglets v0.18):
    * fixed bug that didn't correctly take multiplicity into account for
      the probability of choosing a fraglet for reaction
    * improved debug and execution traces
  20050206  by author and lidia.yamamoto@unibas.ch (fraglets v0.17):
    * added (or recovered?) support for 'nul' operator
    * support for markers /m markername/ in genotype (selfman'05 and wac'05)
    * improved trace printout
  20040704  updated by christian.tschudin@unibas.ch for wac'04 (fraglets v0.15)
  200307??  created by christian.tschudin@unibas.ch (ains'03)
*/

struct frag *fraglist;
struct node *nodelist;
struct name *namelist;
int linecnt;

int debug;
#define DEBUG(l, stmt) if (l <= debug) do { stmt } while (0)
int dump;

extern char *name_lookup(symbol s);

static void f_print(FILE *fout, struct frag *f, int eol);

// ------------------------------------------------------------

int istrans(symbol s)
{
	return s > 0 && s <= SYM_LAST_TRANS;
}

/*
int addhint(struct node *n, symbol s)
{
	DEBUG(9, printf("## addhint on node %s: %s\n", name_lookup(n->name), name_lookup(s)););
	if (n->hintcnt >= MAXHINTS) {
		DEBUG(10, fprintf(stderr, "MAXHINTS (%d) exceeded\n", MAXHINTS););
		return -1;
	}
	n->hints[n->hintcnt] = s;
	n->hintcnt++;

	return 0;
}
*/

// ------------------------------------------------------------

symbol name_add(char *cp)
{
	struct name *n = namelist;
	symbol i = 1;

	while (n && strcmp(n->cp, cp)) {
		n = n->next;
		i++;
	}
	if (!n) {
		n = calloc(1, sizeof(*n));
		n->next = namelist;
		namelist = n;
		n->cp = malloc(strlen(cp) + 1);
		strcpy(n->cp, cp);
		n->sym = i;
	}
	return n->sym;
}

char *name_lookup(symbol s)
{
	struct name *n;

	for (n = namelist; n; n = n->next)
		if (n->sym == s)
			return n->cp;
	return 0;
}


// ------------------------------------------------------------

struct node* node_lookup(symbol node, int create)
{
	struct node *n = nodelist;

	while (n && n->name != node)
		n = n->next;
	if (n)
		return n;
	if (!create)
		return 0;

	n = calloc(1, sizeof(*n));
	n->name = node;
	n->next = nodelist;
	nodelist = n;
	return n;
}

void node_addsegment(struct node *n, symbol segment)
{
	int i;

	if (!n) return;
	DEBUG(9, printf("## addtosegment %s %s\n",
			name_lookup(n->name), name_lookup(segment)););
	for (i = 0; i < MAXSEGMENTS && n->segments[i]; i++);
	if (i < MAXSEGMENTS)
		n->segments[i] = segment;
	else // complain ...
		fprintf(stderr, "no segment place\n");
}

// ------------------------------------------------------------

/* 
struct fragbuf* b_new(int len) // len in bits
{
	struct fragbuf *b;

	b = (struct fragbuf*) malloc(sizeof(struct fragbuf) + (len + 7)/8);
	if (b) {
		b->refcnt = b->len = 0;
	}
	return b;
}
*/

void b_free(struct fragbuf *buf)
{
	free(buf);
}

// ------------------------------------------------------------

struct frag* f_new(struct fragbuf *b, int offs, int len, int mult, symbol n)
{
	struct frag *f = (struct frag*) calloc(1, sizeof(*f));
	if (!f)
		return 0;
	// should check offs and len...
	f->buf = b;
	f->next = 0;
	f->nd = n;
	f->offs = offs;
	f->len = len;
	f->mult = mult;
	b->refcnt++;
	return f;
}

symbol f_peek(struct frag *f)
{
	if (f->offs + 1 > f->buf->len)
		return 0;
	else {
		return f->buf->data[f->offs];
/*	symbol s;
	memcpy(&s, f->buf->data + f->offs, sizeof(s));
	return s;
*/
	}
}


void f_free(struct frag *f)
{
	f->buf->refcnt--;
	if (f->buf->refcnt <= 0)
		b_free(f->buf);
	free(f);
}


void f_freeLinkedList(struct t_queue *q)
{
	struct frag *f;

	while (q->first) {
		f = q->first;
		q->first = f->next;
		f_free(f);
	}
	q->last = 0;
}


int ismatch(symbol s)
{
	return s == SYM_MATCH || s == SYM_MATCHP;
}


void t_append(struct frag *f)
{
	struct node *n = node_lookup(f->nd, 0);
	symbol h = f->buf->data[f->offs];

	DEBUG(9, printf("# append transformation %s\n", name_lookup(h)););

	f->next = 0;
	if (n->tq.last)
		n->tq.last->next = f;
	else
		n->tq.first = f;
	n->tq.last = f;
	//	addhint(n, h);
}


void w_append(struct frag *f)
{
	struct node *n = node_lookup(f->nd, 0);

	DEBUG(9, printf("# append wait %s\n",
			name_lookup(f->buf->data[f->offs])););

	f->next = 0;
	f->expires = n->opcnt + WAIT_DURATION;
	if (n->wait.last)
		n->wait.last->next = f;
	else
		n->wait.first = f;
	n->wait.last = f;
}


void m_add(struct node *n, symbol h, struct frag *f)
{
	DEBUG(9, printf("# adding 'match' for symbol %s\n", name_lookup(h)););

	f->next = 0;
	if (n->mtab[h].last)
		n->mtab[h].last->next = f;
	else
		n->mtab[h].first = f;
	n->mtab[h].last = f;
/*
	if (n->otab[h].first)
		addhint(n, h);
*/
}


void o_add(struct node *n, symbol h, struct frag *f)
{
	DEBUG(9, printf("# adding 'other' for symbol %s\n", name_lookup(h)););

	f->next = 0;
	if (n->otab[h].last)
		n->otab[h].last->next = f;
	else
		n->otab[h].first = f;
	n->otab[h].last = f;
	/*
	if (n->mtab[h].first)
		addhint(n, h);
	*/
}


struct frag* rx_fraglet(struct frag *f) // , int offs, int len)
{
	symbol head;
	struct node *n;

	DEBUG(9, printf("## rx_fraglet\n"););

	if (!n || !f) return 0;

	if (f->len == 0) {
		f_free(f);
		return 0;
	}
	/*
	{
		int i;
		for (i = 0; i < f->len; i++) {
			if (*name_lookup(f->buf->data[f->offs+i]) == 'p') {
				score++;
				break;
			}
		}
	}
	*/
	n = node_lookup(f->nd, 1);

	head = f_peek(f);
	if (istrans(head)) {
		t_append(f);
	} else {
		if (ismatch(head)) {
			DEBUG(9, printf("## match\n"););
			if (f->len == 1) {
				f_free(f);
				return 0;
			}
			head = f->buf->data[f->offs + 1];
			m_add(n, head, f);
		} else
			o_add(n, head, f);
	}
	n->possibleHint |= 1;
	return f;
}


struct frag* f_parsefrag(int lno, char *line)
{
	char *cp, *a, *b;
	int m;
	struct fragbuf *buf;
	symbol n;

	// format: "poolname [ sym1 : sym2 : ... ] mult"

	for (cp = line; isspace(*cp); cp++);
	a = strchr(cp, '[');
	b = strchr(cp, ']');
	if (!a || !b) {
		// this also captures markers of the form:  /m markername/
		return 0;
	}
	buf = calloc(1, sizeof(*buf));
	*a = *b = 0;

	for (line = cp; *cp && !isspace(*cp); cp++);
	*cp = 0;
	n = name_add(line);
	m = atoi(b+1);
	if (m == 0)
	    m = 1;
	if (m <= 0) goto out;
	while (*++a) {
		char c;
		for (; isspace(*a); a++);
		if (!*a)
			break;
		b = a + 1;
		if (*a != '*' && *a != '_') {
			if (!isalnum(*a)) {
				fprintf(stderr, "parse error --- %s\n", a);
				goto out;
			}
			for (b=a+1; isalnum(*b) || *b == '_'; b++);
		}
		if (*b && !isspace(*b) && *b != ':') {
			fprintf(stderr, "parse error --. %s\n", a);
			goto out;
		}
		c = *b;
		*b = 0;
		buf->len++;
		buf = realloc(buf, sizeof(*buf) + buf->len * sizeof(symbol));
		buf->data[buf->len - 1] = name_add(a);
		if (isspace(c)) {
			for (b++; isspace(*b); b++);
			c = *b;
		}
		if (!c)
			break;
		if (c != ':') {
			fprintf(stderr, "-.. %s\n", b);
			goto out;
		}
		a = b;
	}
	return f_new(buf, 0, buf->len, m, n);
 out:
	free(buf);
	return 0;
}


// parse genotype
struct frag* f_parse(int lno, char *line)
{
	char *t;
	struct frag *flist = 0, *f;

	if (*line != '/')
		return 0;

	for (t = strtok(line+1, "/\n"); t ; t = strtok(NULL, "/\n")) {
		f = f_parsefrag(lno, t);
		if (f) {
			f->next = flist;
			flist = f;
		}
	}
	return flist;
}


symbol seg_parse(int lno, char *line, symbol *node)
{
	char *cp;
	symbol s;

	//  format: "A <nodename> <segname>" -- attach

	cp = strtok(line+1, " \t\n");
	*node = name_add(cp);
	cp = strtok(0, " \t\n");
	s = name_add(cp);
	cp = strtok(0, " \t\n");
	if (!*node || !s || cp) {
		fprintf(stderr, "syntax error on line %d\n", lno);
		return 0;
	}
	return s;
}


int f_import(FILE *fin)
{
	char line[10000], *cp;
	struct frag *f, *f2;
	symbol node, seg;

	while (fgets(line, sizeof(line)-1, fin)) {
		linecnt++;
		switch(line[0]) {
		case '\n':
		case '#':
			continue;
		case 'e':
			return 0;
		case 'f':
		case 'F':
			for (cp = line+1; isspace(*cp); cp++);
			f = f_parsefrag(linecnt, cp);
			if (!f) {
			  // Lidia: 'line' is broken at this stage, cannot be printed
				fprintf(stderr, "Line %d: cannot parse fraglet\n", linecnt);
				exit(-1);
			}
			while (f) {
				f2 = f->next;
				f->next = 0;
				rx_fraglet(f);
				f = f2;
			}
			break;
		case 'g':
		case 'G':
			for (cp = line+1; isspace(*cp); cp++);
			f = f_parse(linecnt, cp);
			if (!f) {
				fprintf(stderr, "Line %d: cannot parse genotype\n", linecnt);
				exit(-1);
			}
			while (f) {
				f2 = f->next;
				f->next = 0;
				rx_fraglet(f);
				f = f2;
			}
			return 0;
		case 'a':
		case 'A':
			seg = seg_parse(linecnt, line, &node);
			if (!seg) {
				fprintf(stderr, "Line %d: cannot parse segment %s\n", linecnt, line);
				exit(-1);
			}
			node_addsegment(node_lookup(node, 1), seg);
			break;
		case 'm':
		case 'M':
			break;
		default:
			fprintf(stderr, "Line %d: cannot parse %s\n", linecnt, line);
			exit(-1);
		}
	}
	return -1;
}


void f_compact_list(struct frag *f, struct frag **last)
{
	struct frag *f2, *f3, **fpp;

	while (f) {
		int len = f->len;
		symbol *sp = f->buf->data + f->offs;

		for (fpp = &f->next, f2 = f->next; f2; ) {
			if (len != f2->len ||
			    memcmp(sp, f2->buf->data + f2->offs,
				   sizeof(symbol) * len)) {
				fpp = &f2->next;
				f2 = f2->next;
				continue;
			}
			// same content!
			f->mult += f2->mult;
			f3 = f2->next;
			f_free(f2);
			f2 = *fpp = f3;
		}
		if (!f->next)
			*last = f;
		f = f->next;
	}
}


static void f_print(FILE *fout, struct frag *f, int eol)
{
	int i;

	fprintf(fout, "%s", name_lookup(f->nd));

	for (i = 0; i < f->len; i++) {
		fprintf(fout, i ? ":%s" : "[%s",
		       name_lookup(f->buf->data[f->offs + i]));
	}
	fprintf(fout, eol ? "]\n" : "]");

}


static void f_export_list(FILE *fout, struct node *n, struct frag *f)
{
	for (; f; f = f->next) {
		fprintf(fout, "f ");
		f_print(fout, f, 0);
		fprintf(fout, "%d\n", f->mult);
	}
}


static void f_export_mtab(FILE *fout, struct node *n, struct m_rec *mtab)
{
	int i;

	for (i = 0; i < MAXSYMBOLS; i++, mtab++) {
		f_compact_list(mtab->first, &(mtab->last));
		f_export_list(fout, n, mtab->first);
	}
}


int f_export(FILE *fout)
{
	struct node *n;

	for (n = nodelist; n; n = n->next) {
		int i;
		char *nname = name_lookup(n->name);
		if (n->segments[0]) {
			for (i = 0; n->segments[i]; i++) {
				fprintf(fout, "a %s %s\n", nname,
				       name_lookup(n->segments[i]));
			}
		}
	}

	for (n = nodelist; n; n = n->next) {
		f_compact_list(n->tq.first, &(n->tq.last));
		f_export_list(fout, n, n->tq.first);
		f_compact_list(n->wait.first, &(n->wait.last));
		f_export_list(fout, n, n->wait.first);
		f_export_mtab(fout, n, n->mtab);
		f_export_mtab(fout, n, (struct m_rec*) n->otab);

	}

	return 0;
}


struct frag* f_shrink(struct frag *f, int cnt)
{
	DEBUG(9, printf("# f_shrink %d\n", f->mult););

	if (f->mult == 1) {
		if (f->len <= cnt) {
			f_free(f);
			return 0;
		}
		f->offs += cnt;
		f->len -= cnt;
	} else {
		f->mult--;
		if (f->len > cnt)
			f = f_new(f->buf, f->offs + cnt,
				  f->len - cnt, 1, f->nd);
	}
//	rx_fraglet(n, f);
	return f;
}


struct frag* f_nop(struct frag *f)
{
	node_lookup(f->nd, 0)->symbalance += f->len;

	// Lidia: added execution trace
	DEBUG(1, printf("%05d  ", node_lookup(f->nd, 0)->opcnt);
	      f_print(stdout, f, 0); printf(" --> "););

	f = f_shrink(f, 1);

	if (f) {
		DEBUG(1, f_print(stdout, f, 1););
		rx_fraglet(f);
	} else
		DEBUG(1, printf("[]\n"););

	return f;
}

// Lidia: f_del is the implementation of 'nul' op. as in ains 2003
struct frag* f_del(struct frag *f)
{
	node_lookup(f->nd, 0)->symbalance += f->len;

	DEBUG(1, printf("%05d  ", node_lookup(f->nd, 0)->opcnt);
	      f_print(stdout, f, 0); printf(" --> "););
	DEBUG(1, printf("[]\n"););
	if (--(f->mult) == 0)
		f_free(f);
	else
		t_append(f);
	return 0;
}

struct frag* f_dup(struct frag *f)
{
	struct fragbuf *buf;
	struct frag *f2;

	DEBUG(9, printf("# f_dup %d\n", f->mult););

	DEBUG(1, printf("%05d  ", node_lookup(f->nd, 0)->opcnt);
	      f_print(stdout, f, 0); printf(" --> "););
	if (f->len < 3)
		goto drop;

// dup is symbol count neutral:
// 	node_lookup(f->nd, 0)->symbalance += 0;

	if (f->mult == 1 && f->buf->refcnt == 1) {
		f->buf->data[f->offs] = f->buf->data[f->offs + 1];
		f->buf->data[f->offs + 1] = f->buf->data[f->offs + 2];
		DEBUG(1, f_print(stdout, f, 1););
		rx_fraglet(f);
		return 0;
	}

	buf = malloc(sizeof(*buf) + sizeof(symbol) * f->len);
	buf->len = f->len;
	buf->refcnt = 0;
	memcpy(buf->data + 2, f->buf->data + f->offs + 2,
	       sizeof(symbol) * (f->len - 2));
	buf->data[0] = f->buf->data[f->offs + 1];
	buf->data[1] = f->buf->data[f->offs + 2];
	f2 = f_new(buf, 0, f->len, 1, f->nd);
	DEBUG(1, f_print(stdout, f2, 1););
	rx_fraglet(f2);

	if (--(f->mult) <= 0)
		f_free(f);
	else
		rx_fraglet(f);

	return 0;

drop:
	node_lookup(f->nd, 0)->symbalance += f->len;

	DEBUG(1, printf("[]\n"););
	if (--(f->mult) == 0)
		f_free(f);
	else
		t_append(f);
	return 0;
}


struct frag* f_exch(struct frag *f)
{
	struct fragbuf *buf;
	struct frag *f2;

	DEBUG(9, printf("# f_exch %d\n", f->mult););

	DEBUG(1, printf("%05d  ", node_lookup(f->nd, 0)->opcnt);
	      f_print(stdout, f, 0); printf(" --> "););
	if (f->len < 4)
		goto drop;

	node_lookup(f->nd, 0)->symbalance += 1;

	if (f->mult == 1 && f->buf->refcnt == 1) {
		symbol s;
		f->offs++;
		f->len--;
		s = f->buf->data[f->offs + 1];
		f->buf->data[f->offs + 1] = f->buf->data[f->offs + 2];
		f->buf->data[f->offs + 2] = s;
		DEBUG(1, f_print(stdout, f, 1););
		rx_fraglet(f);
		return 0;
	}

	buf = malloc(sizeof(*buf) + sizeof(symbol) * (f->len - 1));
	buf->len = f->len - 1;
	buf->refcnt = 0;
	memcpy(buf->data + 3, f->buf->data + f->offs + 4,
	       sizeof(symbol) * (f->len - 4));
	buf->data[0] = f->buf->data[f->offs + 1];
	buf->data[1] = f->buf->data[f->offs + 3];
	buf->data[2] = f->buf->data[f->offs + 2];
	f2 = f_new(buf, 0, f->len - 1, 1, f->nd);
	DEBUG(1, f_print(stdout, f2, 1););
	rx_fraglet(f2);

	if (--(f->mult) <= 0)
		f_free(f);
	else
		rx_fraglet(f);
	return 0;

drop:
	node_lookup(f->nd, 0)->symbalance += f->len;

	DEBUG(1, printf("[]\n"););
	if (--(f->mult) == 0)
		f_free(f);
	else
		t_append(f);
	return 0;
}

struct frag* f_newname(struct frag *f)
{
	// XX currently a noop

  // Lidia: changed this to call 'nop' instead of erasing fraglet as before
  // old code that was here went to f_del
  return f_nop(f);
}


struct frag* f_pop(struct frag *f)
{
	DEBUG(9, printf("# f_pop %d\n", f->mult););

	DEBUG(1, printf("%05d  ", node_lookup(f->nd, 0)->opcnt);
	      f_print(stdout, f, 0); printf(" --> "););
	if (f->len < 3)
		goto drop;

	node_lookup(f->nd, 0)->symbalance += 2;

	if (f->mult == 1 && f->buf->refcnt == 1) {
		f->offs += 2;
		f->len -= 2;
		f->buf->data[f->offs] = f->buf->data[f->offs - 1];
	} else {
		struct fragbuf *b2;
		struct frag *f2;

		b2 = malloc(sizeof(*b2) + sizeof(symbol) * (f->len - 2));
		b2->len = f->len - 2;
		b2->refcnt = 0;
		memcpy(b2->data + 1, f->buf->data + f->offs + 3,
		       sizeof(symbol) * (f->len - 3));
		b2->data[0] = f->buf->data[f->offs + 1];
		f2 = f_new(b2, 0, f->len - 2, 1, f->nd);
		if (--(f->mult) <= 0)
			f_free(f);
		else {
			t_append(f);
//			addhint(node_lookup(f->nd, 0), SYM_POP);
		}
		f = f2;
	}
	DEBUG(1, f_print(stdout, f, 1););
	rx_fraglet(f);
	return f;

drop:
	node_lookup(f->nd, 0)->symbalance += f->len;

	DEBUG(1, printf("[]\n"););
	if (--(f->mult) <= 0)
		f_free(f);
	else
		t_append(f);
	return 0;
}


struct frag* f_split(struct frag *f)
{
	symbol *sp;
	int i;
	struct frag *f2;

	DEBUG(9, printf("# f_split %d\n", f->mult););
	DEBUG(1, printf("%05d  ", node_lookup(f->nd, 0)->opcnt);
	      f_print(stdout, f, 0); printf(" --> "););

	for (i = 1, sp = f->buf->data + f->offs + 1; i < f->len; i++, sp++) {
		if (*sp == SYM_STAR)
			break;
	}
	if (i == 1) { // empty first half
		node_lookup(f->nd, 0)->symbalance += 2;
		f = f_shrink(f, 2);
		DEBUG(1, if (f) f_print(stdout, f, 1); else printf("[]\n"););

		rx_fraglet(f);
		return f;
	}
	if (i >= f->len) { // no star (empty second half)
		node_lookup(f->nd, 0)->symbalance += 1;
		f = f_shrink(f, 1);
		DEBUG(1, if (f) f_print(stdout, f, 1); else printf("[]\n"););
		rx_fraglet(f);
		return f;
	}
	node_lookup(f->nd, 0)->symbalance += 2;
	if (i != (f->len - 1)) { // accept the second half, if not empty
		f2 = f_new(f->buf, f->offs + i + 1, f->len - i - 1, 1, f->nd);
		DEBUG(1, f_print(stdout, f2, 0); printf(", "););
		rx_fraglet(f2);
	}
	// handle the first half
	if (f->mult == 1) {
		f->offs += 1;
		f->len = i - 1;
		DEBUG(1, f_print(stdout, f, 1););
	} else {
		f->mult--;
		f2 = f_new(f->buf, f->offs + 1, i - 1, 1, f->nd);
		DEBUG(1, f_print(stdout, f2, 1););
		rx_fraglet(f2);
	}
	rx_fraglet(f);
	return 0;
}


struct frag* f_send(struct frag *f)
{
	symbol seg;
	struct node *dest, *n;
	struct frag *f2;
	int i;

	DEBUG(1, printf("%05d  ", node_lookup(f->nd, 0)->opcnt);
	      f_print(stdout, f, 0); printf(" --> "););
 	if (f->len < 4)
		goto drop;
	
	seg = f->buf->data[f->offs + 1];

	dest = node_lookup(f->buf->data[f->offs + 2], 0);
	if (!dest)
		goto drop;

	DEBUG(9, printf("## send from %s to %s (%d)\n",
			name_lookup(f->nd), name_lookup(dest->name),
			f->mult););

	n = node_lookup(f->nd, 0);
	for (i = 0; i < MAXSEGMENTS && n->segments[i]; i++)
		if (n->segments[i] == seg)
			break;
	if (i >= MAXSEGMENTS || !n->segments[i])
		goto drop;

	for (i = 0; i < MAXSEGMENTS && dest->segments[i]; i++)
		if (dest->segments[i] == seg)
			break;
	if (i >= MAXSEGMENTS || !dest->segments[i])
		goto drop;

	node_lookup(f->nd, 0)->symbalance -= f->len + 3;
	node_lookup(dest->name, 0)->symbalance += f->len - 3;

	if (f->mult == 1) {
		f->offs += 3;
		f->len -= 3;
		f2 = f;
		f2->nd = dest->name;
	} else {
		f->mult--;
		f2 = f_new(f->buf, f->offs + 3, f->len - 3, 1, dest->name);
		t_append(f);
//		addhint(node_lookup(f->nd, 0), SYM_SEND);
	}
	DEBUG(1, f_print(stdout, f2, 1););
	rx_fraglet(f2);

	return f;
drop:
	node_lookup(f->nd, 0)->symbalance += f->len;

	DEBUG(1, printf("[]\n"););
	if (--(f->mult) == 0)
		f_free(f);
	else
		t_append(f);
	return 0;
}


struct frag* f_wait(struct frag *f)
{
	DEBUG(1, printf("%05d  ", node_lookup(f->nd, 0)->opcnt);
	      f_print(stdout, f, 0); printf(" --> ...\n"););

	f = f_shrink(f, 0);
	if (f)
		w_append(f);
	return f;
}


struct frag* f_pick(struct t_queue *q, int offs)
{
	struct frag *f, *prev;

/*
	printf("f_pick %d\n", offs);
	printf("\n* - - - - - - - - - - -\n");
	f_export(stdout);
	printf("\n* - - - - - - - - - - -\n");
*/


	// pick the fraglet at offset 'offs', taking its
	// multiplicity into account
	prev = 0;
	for (f = q->first; f; f = f->next) {
//		printf("f_pick mult=%d\n", f->mult);
		if (offs < f->mult)
			break;
		offs -= f->mult;
		prev = f;
	}
	if (!f) {
		fprintf(stderr, "# f_pick: no such fraglet\n");
		return 0;
	}
//	printf("f_pick found %d, next=%p\n", offs, f->next);

	// remove the fraglet from the queue
	if (!prev) {
		q->first = f->next;
	} else {
		prev->next = f->next;
	}
	if (q->last == f)
		q->last = prev;
	f->next = 0;

	return f;
}

int f_execFromQueue(struct node *n, struct t_queue *q, int offs)
{
	struct fragbuf *buf;
	struct frag *f, *prev, *f1, *f2, *f3;
	int ops = 0;
	int s;

	DEBUG(9, printf("## f_execFromQueue %d\n", offs););

	f = f_pick(q, offs);
	if (!f)
		return 0;

//	printf("f_exec mult=%d\n", f->mult);

	s = f_peek(f);
	if (istrans(s)) {
/*
		struct frag *f = n->tq.first;
		if (!f)
			return ops;
		n->tq.first = f->next;
		if (n->tq.last == f)
			n->tq.last = 0;
		f->next = 0;
		s = f_peek(f);
*/
		switch (s) {
		case SYM_NOP:   f_nop(f);       break;

		// Lidia: added SYM_DEL for 'nul' op as in ains2003
		case SYM_DEL:   f_del(f);   break;

		case SYM_DUP:   f_dup(f);       break;
		case SYM_EXCH:  f_exch(f);      break;
		case SYM_NEW:   f_newname(f);   break;
		case SYM_POP:	f_pop(f);       break;
		case SYM_SEND:  f_send(f);      break;
		case SYM_SPLIT: f_split(f);     break;
		case SYM_WAIT:  f_wait(f);     break;
		case SYM_TRIGGER: appl_trigger(f); break;
		default:
			fprintf(stderr, "'%s' not implemented\n",
			       name_lookup(s));
			if (--(f->mult) == 0)
				f_free(f);
			break;
		}
		return ops+1;
	}

//	printf("reaction ...\n");
	if (ismatch(s)) {
		struct o_rec *or;
		int i;
		f1 = f;
		or = n->otab + f->buf->data[f->offs + 1];
		i = random() % or->active_cnt;
/*
		printf("found f1, now picking match %d from %d\n",
		       i, or->active_cnt);
*/
		f2 = f_pick((struct t_queue*) or, i);
		if (!f2) {
			rx_fraglet(f1);
			return 0;
		}
	} else {
		struct m_rec *mr;
		int i;

		f2 = f;
		mr = n->mtab + s;
		i = random() % mr->active_cnt;
/*
		printf("found f2, now picking match %d from %d\n",
		       i, mr->active_cnt);
*/
		f1 = f_pick((struct t_queue*) mr, i);
		if (!f1) {
			rx_fraglet(f2);
			return 0;
		}
	}

/*
	f1 = n->mtab[s].first;
	f2 = n->otab[s].first;
	if (!f1 || !f2)
		return ops;
*/
	DEBUG(1, printf("%05d  ", n->opcnt);
	      f_print(stdout, f1, 0); printf(", ");
	      f_print(stdout, f2, 0); printf(" --> "););
	if ((f1->len + f2->len - 3) > 0) {
		buf = malloc(sizeof(*buf) + sizeof(symbol) * (f1->len + f2->len - 3));
		buf->refcnt = 0;
		buf->len = f1->len + f2->len - 3;
		memcpy(buf->data, f1->buf->data + f1->offs + 2, sizeof(symbol)*(f1->len - 2));
		memcpy(buf->data + f1->len - 2, f2->buf->data + f2->offs + 1, sizeof(symbol)*(f2->len - 1));
		f3 = f_new(buf, 0, buf->len, 1, f1->nd);
		switch (f1->buf->data[f1->offs]) {
		case SYM_MATCH:
			DEBUG(1, f_print(stdout, f3, 1););
			break;
		case SYM_MATCHP:
			DEBUG(1, f_print(stdout, f1, 0);
			      fprintf(stdout, ", ");
			      f_print(stdout, f3, 1););
			break;
		default:
			fprintf(stderr, "wrong MATCH fragment in list\n");
			break;
		}
		node_lookup(f3->nd, 0)->symbalance -= f3->len;
		rx_fraglet(f3);
	} else {
		switch (f1->buf->data[f1->offs]) {
		case SYM_MATCH:
			DEBUG(1, fprintf(stdout, "[]\n"););
			break;
		case SYM_MATCHP:
			DEBUG(1, f_print(stdout, f1, 0);
			      fprintf(stdout, ", []\n"););
			break;
		default:
			fprintf(stderr, "wrong MATCH fragment in list\n");
			break;
		}
	}

/*
	n->mtab[s].first = f1->next;
	if (n->mtab[s].last == f1)
		n->mtab[s].last = 0;
	n->otab[s].first = f2->next;
	if (n->otab[s].last == f2)
		n->otab[s].last = 0;
*/

	if (f1->buf->data[f1->offs] == SYM_MATCH) {
		node_lookup(f1->nd, 0)->symbalance += f1->len;
		f1->mult--;
	} // else SYM_MATCHP
	if (f1->mult <= 0)
		f_free(f1);
	else
		rx_fraglet(f1);

	node_lookup(f2->nd, 0)->symbalance += f2->len;
	f2->mult--;
	if (f2->mult <= 0)
		f_free(f2);
	else
		rx_fraglet(f2);

/*
	printf("\n# - - - - - - - - - - -\n");
	f_export(stdout);
	printf("\n# - - - - - - - - - - -\n");
*/

	return ops+1;
	
}


int f_exec(struct node *n, symbol s, int offs)
{
	struct fragbuf *buf;
	struct frag *f1, *f2, *f3;
	int ops = 0;

	DEBUG(9, printf("## f_exec %s\n", name_lookup(s)););

	if (istrans(s)) {
		struct frag *f = n->tq.first;
		if (!f)
			return ops;
		n->tq.first = f->next;
		if (n->tq.last == f)
			n->tq.last = 0;
		f->next = 0;
		s = f_peek(f);
		switch (s) {
		case SYM_NOP:   f_nop(f);       break;

		// Lidia: added SYM_DEL for 'nul' op as in ains2003
		case SYM_DEL:   f_del(f);   break;

		case SYM_DUP:   f_dup(f);       break;
		case SYM_EXCH:  f_exch(f);      break;
		case SYM_NEW:   f_newname(f);   break;
		case SYM_POP:	f_pop(f);       break;
		case SYM_SEND:  f_send(f);      break;
		case SYM_SPLIT: f_split(f);     break;
		case SYM_WAIT:  f_wait(f);     break;
		case SYM_TRIGGER: appl_trigger(f); break;
		default:
			fprintf(stderr, "'%s' not implemented\n",
			       name_lookup(s));
			if (--(f->mult) == 0)
				f_free(f);
			break;
		}
		return ops+1;
	}
	f1 = n->mtab[s].first;
	f2 = n->otab[s].first;
	if (!f1 || !f2)
		return ops;
	DEBUG(1, printf("%05d  ", n->opcnt);
	      f_print(stdout, f1, 0); printf(", ");
	      f_print(stdout, f2, 0); printf(" --> "););
	if ((f1->len + f2->len - 3) > 0) {
		buf = malloc(sizeof(*buf) + sizeof(symbol) * (f1->len + f2->len - 3));
		buf->refcnt = 0;
		buf->len = f1->len + f2->len - 3;
		memcpy(buf->data, f1->buf->data + f1->offs + 2, sizeof(symbol)*(f1->len - 2));
		memcpy(buf->data + f1->len - 2, f2->buf->data + f2->offs + 1, sizeof(symbol)*(f2->len - 1));
		f3 = f_new(buf, 0, buf->len, 1, f1->nd);
		switch (f1->buf->data[f1->offs]) {
		case SYM_MATCH:
			DEBUG(1, f_print(stdout, f3, 1););
			break;
		case SYM_MATCHP:
			DEBUG(1, f_print(stdout, f1, 0);
			      fprintf(stdout, ", ");
			      f_print(stdout, f3, 1););
			break;
		default:
			fprintf(stderr, "wrong MATCH fragment in list\n");
			break;
		}
		node_lookup(f3->nd, 0)->symbalance -= f3->len;
		rx_fraglet(f3);
	} else {
		switch (f1->buf->data[f1->offs]) {
		case SYM_MATCH:
			DEBUG(1, fprintf(stdout, "[]\n"););
			break;
		case SYM_MATCHP:
			DEBUG(1, f_print(stdout, f1, 0);
			      fprintf(stdout, ", []\n"););
			break;
		default:
			fprintf(stderr, "wrong MATCH fragment in list\n");
			break;
		}
	}

	n->mtab[s].first = f1->next;
	if (n->mtab[s].last == f1)
		n->mtab[s].last = 0;
	n->otab[s].first = f2->next;
	if (n->otab[s].last == f2)
		n->otab[s].last = 0;

// we could do better here: append directly, we know the front symbol s ...
	if (f1->buf->data[f1->offs] == SYM_MATCH) {
		node_lookup(f1->nd, 0)->symbalance += f1->len;
		f1->mult--;
	}
	if (f1->mult <= 0)
		f_free(f1);
	else
		rx_fraglet(f1);

	node_lookup(f2->nd, 0)->symbalance += f2->len;
	f2->mult--;
	if (f2->mult <= 0)
		f_free(f2);
	else
		rx_fraglet(f2);

	return ops+1;
}


void f_cleanup()
{
	struct node *n;

	DEBUG(9, printf("## cleanup\n"););
	for (n = nodelist; n; n = n->next) {
		int i;
		struct m_rec *mtab;
		struct o_rec *otab;

		for (i = 0, mtab = n->mtab; i < MAXSYMBOLS; i++, mtab++)
			f_freeLinkedList((struct t_queue *) mtab);
		for (i = 0, otab = n->otab; i < MAXSYMBOLS; i++, otab++)
			f_freeLinkedList((struct t_queue *) otab);
		f_freeLinkedList(&(n->tq));

		n->opcnt = 0;
		n->r_active = n->t_active = 0;
		//		n->hintcnt = 0;
	}

}


void collect_hints(struct node *n)
{
	int i;
	struct m_rec *mtab;
	struct o_rec *otab;
	struct frag *f;

	DEBUG(9, printf("## collecthints %s\n", name_lookup(n->name)););

	/*
	if (n->tq.first) {
		if (addhint(n, f_peek(n->tq.first)))
			return;
	}
	for (i = 0, mtab = n->mtab; i < MAXSYMBOLS; i++, mtab++)
		if (mtab->first && n->otab[i].first) {
			if (addhint(n, i))
				return;
		}
	*/
	n->r_active = n->t_active = 0;
	for (i = 0, mtab = n->mtab, otab = n->otab;
				     i < MAXSYMBOLS; i++, mtab++, otab++) {
		mtab->active_cnt = 0;
		for (f = mtab->first; f; f = f->next) {
			if (otab->first)
				mtab->active_cnt += f->mult;
		}

		otab->active_cnt = 0;
		for (f = otab->first; f; f = f->next) {
			if (mtab->first)
				otab->active_cnt += f->mult;
		}
		mtab->r_active_cnt = mtab->active_cnt;
		if (mtab->active_cnt > otab->active_cnt)
			mtab->r_active_cnt = otab->active_cnt;

		n->r_active += mtab->r_active_cnt;
//		n->o_active += mtab->r_active_cnt;
	}
	for (f = n->tq.first; f; f = f->next)
		n->t_active += f->mult;
	n->total_active = n->r_active + n->t_active;
}


void run(int lim)
{
	int i;
	struct node *n;
	symbol s;
	int clock = 0;

	while (lim > clock) {
		DEBUG(9, printf("## run loop %d/%d\n", clock, lim););
		for (n = nodelist; n; n = n->next) {
			if (n->wait.first && n->opcnt >= n->wait.first->expires) {
				struct frag *f = n->wait.first;
				n->wait.first = f->next;
				if (n->wait.last == f)
					n->wait.last = 0;
				f->next = 0;
				f = f_shrink(f, 1);
				if (f)
					rx_fraglet(f);
				n->possibleHint |= 1;
			}
			n->possibleHint |= appl_clockEvent(n, clock);
			/*
			if (!n->hintcnt && n->possibleHint)
				collect_hints(n);
			if (n->hintcnt) {
				i = random() % n->hintcnt;
				s = n->hints[i];
				n->hintcnt--;
				if (i != n->hintcnt)
					n->hints[i] = n->hints[n->hintcnt];
				i = f_exec(n, s);
			} else
				n->possibleHint = 0;
			*/
			collect_hints(n);
			if (n->total_active) {
				i = random() % n->total_active;
//				printf("i=%d from %d\n", i, n->total_active);
				if (i < n->r_active) {
					struct m_rec *mr;
					mr = n->mtab;
					while (i >= mr->r_active_cnt) {
//						printf("m\n");
						i -= mr->r_active_cnt;
						mr++;
					}
//					printf("picking mtab %d\n", i);
					i = random() % mr->active_cnt;
					f_execFromQueue(n, (struct t_queue *)mr, i);
				} else {
						i -= n->r_active;
						f_execFromQueue(n, &n->tq, i);
				}
//				}
			}
			n->opcnt++;
		}
		clock++;
	}
}

int main(int argc, char *argv[])
{
	int lim = 100;
	int soffs = 0;

	for (argc--, argv++; argc > 0; argc--, argv++) {
		if (!strcmp(argv[0], "-d") && argc > 1) {
			debug = atoi(argv[1]);
			argc--;
			argv++;
			continue;
		}
		if (!strcmp(argv[0], "-e") && argc > 1) {
			dump = atoi(argv[1]);
			argc--;
			argv++;
			continue;
		}
		if (!strcmp(argv[0], "-lim") && argc > 1) {
			lim = atoi(argv[1]);
			argc--;
			argv++;
			continue;
		}
		if (!strcmp(argv[0], "-soffs") && argc > 1) {
			soffs = atoi(argv[1]);
			argc--;
			argv++;
			continue;
		}
		fprintf(stderr, "unknown option %s\n", argv[0]);
		fprintf(stderr, "usage: frag [options]\n");
		fprintf(stderr, "  -d lev     // debug\n");
		fprintf(stderr, "  -e N       // export: 1=in, 2=out, 3=both\n");
		fprintf(stderr, "  -lim N     // limit exec count\n");
		fprintf(stderr, "  -soffs N   // score index offset\n");
		exit(-1);
	}

	// (lidia) be careful to add symbols in the same order as frag.h
	// enum SYM_*

	name_add("nop");
	// Lidia: added SYM_DEL="nul"
	name_add("nul");
	name_add("dup");
	name_add("exch");
	name_add("new");
	name_add("pop");
	name_add("send");
	name_add("split");
	name_add("wait");
	name_add("_");

	name_add("match");
	name_add("matchp");
	name_add("*");
	// SYM_NUL is end-of-list marker

	while (f_import(stdin) == 0) {

		appl_scoreReset();
		// Lidia: reset also seed for random(), for repeatability
		srandom(1);

		if (dump || debug) {
			time_t t;
			struct passwd *u;
			struct utsname un;

			time(&t);
			u = getpwuid(getuid());
			uname(&un);

			printf("# " VERSION);
			printf("# Executed by %s on %s at %s",
			       u->pw_name, un.nodename, ctime(&t));
		}

		if (dump == 1 || dump == 3) {
			printf("\n# - - - - - - - - - - -\n");
			f_export(stdout);
		}
		if (debug)
			printf("\n# - - - - - - - - - - -\n");

		run(lim);

		if (dump == 2 || dump == 3) {
			printf("\n# - - - - - - - - - - -\n");
			f_export(stdout);
		}

		if (lim > 0) {
			if (dump || debug)
				printf("\n# - - - - - - - - - - -\n");
			//	printf("# score: %d\n", finalScore());
			printf("s %d %d\n", appl_finalScore(), ++soffs);
		}

		f_cleanup();
	}

	return 0;
}

// eof
