/* fraglets0.17/src/mutate.c

  Copyright (C) 2002-2005 Christian Tschudin <tschudin@docs.uu.se>

  This file is distributed under the terms of the GNU general Public
  License (GPL), see the file COPYING

  History:
  20050131 bug fixes by lidia.yamamamoto@unibas.ch (fraglets v0.17)
  20040523 updated (fraglets v0.15)
  2002???? created by tschudin@docs.uu.se

*/


/*
  mutates a system description (genotype) once.

  Function: picks a patchable place in the genotype with
  equal probability (either a symbol, or the multiplicity count),
  and mutates that place. Either by replacing the symbol or by
  increasing/decreasing the count by one, respectively.

Input:

  line 1:  system description e.g.
    g /a[match:b:nothing]1;a[split:a:*:hello]2/b[_:aha]1;b[split:a:*]2/

  line 2 and higher: weight of replacement symbols, e.g.
    10 newsym
    10 othersym
    10 differentsym
    70 $15

  (the $15 format says: put a tag with number in the range 0..14, e.g. t11)
  In this case: in 7 of 20 cases we insert a tag, in 10 of 20 cases we
  just change the fraglet count.

*/

#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <string.h>
#include <sys/timeb.h>

struct mut {
	char *cp;
	int w;
	struct mut *next;
};


int main(int argc, char *argv[])
{
	char g[5000], s[30], *cp, *startfrag;
	int symcnt, tot, r;
	int i;
	struct mut *mutlist = 0, *m;

/*	int debug = 0;
	if (argv[1] && !strcmp(argv[1], "-d"))
		debug = 1;
*/

	{
		struct timeb t;

		ftime(&t);
		srandom(t.time ^ t.millitm );
	}

	do {
		if (!fgets(g, sizeof(g)-1, stdin))
			return -1;
		if (g[0] == 'g')
			break;
	} while (1);
			
	tot = 0;
	while (fgets(s, sizeof(s)-1, stdin)) {
		if (2 != sscanf(s, "%d %25s\n", &i, &s))
			continue;
		m = malloc(sizeof(*m));
		m->w = i;
		tot += i;
		m->cp = malloc(strlen(s) + 1);
		strcpy(m->cp, s);
		m->next = mutlist;
		mutlist = m;
	}
	if (tot == 0)
		return -1;

	// count places where a modifiable symbol/fragletcount starts
	cp = strchr(g, '/');
	if (!cp)
		return -1;
	for (symcnt = 0; *cp; cp++)
		if (*cp == '[' || *cp == ':' || *cp == ']')
			symcnt ++;

	// locate the place to patch
	r = random() % symcnt;
	//fprintf(stderr, "r=%d,  symcnt=%d\n", r, symcnt);

	for (i = -1, cp = g; i < r && *cp; cp++) {
	  // Lidia: mark start of fraglet to patch
	  if (*cp == '/')
	    startfrag = cp;
	  if (*cp == '[' || *cp == ':' || *cp == ']')
	    i++;
	}

	// Lidia: safety test
	if (!*cp) {
	  fprintf(stderr,"mutate: wrong genotype format\n");
	  exit(-1);
	}

	//printf("startfrag=<%s> rest=<%s>\n", startfrag, cp);

	if (isdigit(*cp)) { // patch multiplicity or number symbol
	  // (lidia) bug: doesn't recognize negative numbers, and will
	  // try to numerically patch symbols such as '1b3' '2ab'
		i = atoi(cp);
		if (random() % 2)
			i++;
		else
			i--;
		*cp = 0;
		// find end of patch zone
		// (lidia): if numeric symbol then delimiter is ':'
		//   if multiplicity then delimiter is '/'
		for (cp++; *cp && *cp != ':' && *cp != '/'; cp++);

		// Lidia: if mult<=0, remove frag
		if (i<=0 && *cp == '/') {
		  *startfrag = 0;
		  printf("%s%s", g, cp);
		} else {
		  // insert multiplicity patch
		  printf("%s%d%s", g, i, cp);
		}
	} else { // patch symbol
		// choose replacement symbol
		r = random() % tot;
//		printf("r = %d / %d\n", r, tot);
		for (m = mutlist; r > m->w; m = m->next) {
//			printf("  %d %s\n", m->w, m->cp);
			r -= m->w;
		}
//		printf("* %d %s\n", m->w, m->cp);

		*cp = 0;
		// find end of patch zone
		for (cp++; *cp && *cp != ':' && *cp != ']'; cp++);
		if (m->cp[0] == '$') // insert tag with number
			printf("%st%d%s", g,
			       (int)(random() % atoi(m->cp+1)), cp);
		else // insert patch
			printf("%s%s%s", g, m->cp, cp);
	}
	return 0;
}
